module Main (main) where

import Control.Concurrent.STM
import Data.List (intersperse)
import Data.Foldable (toList)
import Quasar
import Quasar.Prelude
import Quasar.Timer
import Quasar.Wayland.Client
import Quasar.Wayland.Client.Buffer
import Quasar.Wayland.Client.JuicyPixels
import Quasar.Wayland.Protocol
import Quasar.Wayland.Protocol.Generated

import Codec.Picture

import Glest.Egl
import Glest.Renderer


data Dimensions = Dimensions {
  width :: Int,
  height :: Int,
  aspect :: Double
}

data Position = Position {
  dimensions :: Dimensions,
  pixelX :: Int,
  pixelY :: Int,
  u :: Double,
  v :: Double,
  x :: Double,
  y :: Double
}

mkDimensions :: Int -> Int -> Dimensions
mkDimensions width height = Dimensions { width, height, aspect }
  where
    aspect :: Double
    aspect = (fromIntegral width) / (fromIntegral height)

mkPosition :: Dimensions -> Int -> Int -> Position
mkPosition dimensions pixelX pixelY = Position { dimensions, pixelX, pixelY, u, v, x, y }
  where
    width' = width dimensions
    height' = height dimensions
    u :: Double
    u = (fromIntegral pixelX) / (fromIntegral width')
    v :: Double
    v = (fromIntegral pixelY) / (fromIntegral height')
    innerRadius :: Int
    innerRadius = div (min width' height') 2
    x :: Double
    x = (fromIntegral $ pixelX - (div width' 2)) / (fromIntegral innerRadius)
    y :: Double
    y = (fromIntegral $ pixelY - (div height' 2)) / (fromIntegral innerRadius)

color :: RealFrac a => a -> a -> a -> PixelRGBA8
color r g b = PixelRGBA8 (toWord r) (toWord g) (toWord b) 255
  where
    toWord = truncate . (* 255) . (max 0) . (min 1)

wallpaper :: Position -> PixelRGBA8
wallpaper p = color (u p) 0 (v p)

wallpaper2 :: Position -> PixelRGBA8
wallpaper2 p = color 255 0 0

wallpaperImage :: (Position -> PixelRGBA8) -> Image PixelRGBA8
wallpaperImage fn = generateImage pixel width height
  where
    width :: Int
    width = 1920
    height :: Int
    height = 1080
    dimensions :: Dimensions
    dimensions = mkDimensions width height
    pixel :: Int -> Int -> PixelRGBA8
    pixel x y = fn $ mkPosition dimensions x y

--generateWallpaper :: IO ()
--generateWallpaper = writePng "/tmp/wallpaper.png" wallpaperImage


main :: IO ()
main = do
  dmabuf <- renderDmabufTest
  waylandTest dmabuf

waylandTest :: Dmabuf -> IO ()
waylandTest dmabuf = do
  withRootResourceManager do
    traceIO "Connecting"
    client <- connectWaylandClient
    traceIO "Connected"

    join $ liftIO $ atomically do

      wlCompositor <- bindSingleton @Interface_wl_compositor client.registry

      shm <- newShmBufferManager client

      wlSurface <- wlCompositor.create_surface
      setMessageHandler wlSurface EventHandler_wl_surface {
        enter = \_ -> pure (),
        leave = \_ -> pure ()
      }

      --xdgWmBase <- bindSingleton @Interface_xdg_wm_base client.registry
      --setMessageHandler xdgWmBase EventHandler_xdg_wm_base {
      --  ping = \serial -> xdgWmBase.pong serial
      --}

      --xdgSurface <- xdgWmBase.get_xdg_surface wlSurface
      --setMessageHandler xdgSurface EventHandler_xdg_surface {
      --  configure = \serial -> xdgSurface.ack_configure serial
      --}
      --xdgToplevel <- xdgSurface.get_toplevel
      --setMessageHandler xdgToplevel EventHandler_xdg_toplevel {
      --  configure = \_ _ _ -> pure (),
      --  close = pure ()
      --}

      --xdgToplevel.set_title "foobar"


      wlrLayerShell <- bindSingleton @Interface_zwlr_layer_shell_v1 client.registry

      configuredVar <- newTVar False

      wlrLayerSurface <- wlrLayerShell.get_layer_surface wlSurface Nothing 2 "demo"
      setMessageHandler wlrLayerSurface EventHandler_zwlr_layer_surface_v1 {
        configure = \serial width height -> do
          wlrLayerSurface.ack_configure serial
          writeTVar configuredVar True,
        closed = pure ()
      }
      wlrLayerSurface.set_size 512 512

      wlSurface.commit
      -- Must await first surface `configure` event

      zwpLinuxDmabuf <- bindSingleton @Interface_zwp_linux_dmabuf_v1 client.registry
      bufferParams <- zwpLinuxDmabuf.create_params
      setMessageHandler bufferParams EventHandler_zwp_linux_buffer_params_v1 {
        created = \_ -> traceM "Error: Buffer created even though create_immed was used",
        failed = traceM "Buffer creation failed"
      }

      forM (zip [0,1..] dmabuf.planes) \(i, plane) ->
        --bufferParams.add plane.fd i plane.offset plane.stride 0x00ffffff 0xffffffff
         bufferParams.add plane.fd i plane.offset plane.stride (modifierHi dmabuf) (modifierLo dmabuf)

      buffer <- bufferParams.create_immed 512 512 dmabuf.fourcc 0

      pure do
        --buffer <- liftIO $ loadImageBuffer shm "cat.png"
        --buffer <- liftIO $ toImageBuffer shm (wallpaperImage wallpaper)

        liftIO $ atomically do
          check =<< readTVar configuredVar
          wlSurface.attach (Just buffer) 0 0
          wlSurface.commit

        --buffer2 <- liftIO $ toImageBuffer shm (wallpaperImage wallpaper)

        --liftIO $ atomically do
        --  wlSurface.attach (Nothing) 0 0
        --  --wlSurface.damage 0 0 100 100
        --  wlSurface.commit

        --traceIO . ("shm buffer formats: " <>) . mconcat . intersperse ", " . fmap show . toList =<< await shm.formats

        pure ()

    traceIO "Waiting..."
    await =<< newDelay 2_000_000
    traceIO "Closing"
  traceIO "Closed"
